{ TLinks  Version 3.1  Copyright 1996 Alexander Petrosyan  }
{                                      2:5020/468.8         }
{                                      (095)535-2222        }

unit DsnLinks;

interface

uses
  SysUtils, WinTypes, WinProcs, Messages, Classes, Graphics, Controls,
  Forms, ExtCtrls;

type
  TCircleSize = 0..10;  { ,   -- property editor}
  TLineWidth = 1..7;
  THintPause = 1..10000;

const
  DefaultCircleSize = 3;
  DefaultLineWidth = 1;
  DefaultMasterColor = clRed;
  DefaultSelfColor = clWhite;
  DefaultSlaveColor = clNavy;
  DefaultHintColor = $80FFFF;  { default hint window color }
  DefaultHintPause = 200;
  DefaultMasterProps =
    'MasterSource;DataSource;DataSet '+
    'Source;Destination '+
    'DdeConv '+

    {Orpheus}
    'Controller '+

    '';
  DefaultSlaveProps =
    'FocusControl '+
    'PopupMenu '+
    'LookupSource;LookupDisplay '+
    'DirLabel;DirList;FileEdit;FileList '+

    {Orpheus}
    'FocusedControl '+

    '';
  DefaultHiddenProps =
    'Name '+
    'Left;Top;Width;Height '+
    'Caption '+
    'TabOrder '+
    'IsControl '+
    '';

const
  ListDelims = [' ', ',', ';'];

type
  TLinks = class
  private
    { Private declarations }
    Demo: Boolean;

  public
    { Public declarations }
    constructor Create;
    destructor Destroy; override;

  private
    { Private declarations }
    function ComponentRect(Component: TComponent): TRect;
    function ComponentBelongsToForm(Component: TComponent): Boolean;

  private
    { Private declarations }
    procedure Update;

  private
    { Private declarations }
    FTemporaryDisabled: Boolean;

    procedure SetTemporaryDisabled(Value: Boolean);
    property TemporaryDisabled: Boolean read FTemporaryDisabled write SetTemporaryDisabled;

  private
    { Private declarations }
    MouseHook, KeyboardHook: HHook;

    function WindowHook(var Message: TMessage): Boolean; export;

    procedure HookUnhookMainWindow(Hook: TWindowHook; wParam: Word);
    procedure Hook;
    procedure UnHook;

  private
    { Private declarations }
    Timer: TTimer;

    procedure TimerElapsed(Sender: TObject);

  private
    { Private declarations }
    SavedScreen: TBitmap;
    SavedRectangle: TRect;
    HintForm: TForm;
    HintComponent: TComponent;

    procedure ActivateHint(AHintComponent: TComponent);
    procedure Cancel;

    procedure CollectLinks(Structure: TStrings; Component: TComponent);
    function GetHint(Component: TComponent): PChar;

  private
    { Private declarations }
    Active: Boolean;
    CircleSize: TCircleSize;
    LineWidth: TLineWidth;
    MasterColor, SelfColor, SlaveColor: TColor;
    HintColor: TColor;
    MasterProps, SlaveProps,
    HiddenProps: string;
    ShowParentRect, ShowSelfRect, ShowLinks, ShowHint: Boolean;

    PrevIniFileAge: Longint;

    procedure StoreOptions;
    procedure LoadOptions;
  end;

var
  Links: TLinks{   ,    impl?};

procedure Register;

implementation

uses
  TypInfo,
  Consts,
  IniFiles;

const
  IniFileName = 'Delphi.INI';
  SectionName = 'TLinks';
  IMasterProps = 'MasterProps';
  ISlaveProps = 'SlaveProps';
  IHiddenProps = 'HiddenProps';
  ICircleSize = 'CircleSize';
  ILineWidth = 'LineWidth';
  IMasterColor = 'MasterColor';
  ISelfColor = 'SelfColor';
  ISlaveColor = 'SlaveColor';
  IHintColor = 'HintColor';
  IHintPause = 'HintPause';
  IShowParentRect = 'ShowParentRect';
  IShowSelfRect = 'ShowSelfRect';
  IShowLinks = 'ShowLinks';
  IShowHint = 'ShowHint';

const
  IconSize = 24 +2{}*2{  };
const
  SMaster = '!' { ,   };
  SSlave = '' { __ - ,    =};


procedure Register;  begin end;


  {--------------- Word manipulation -------------------------------}
  type
    CharSet = set of Char;

  function ExtractWord(N : Byte; S : string; WordDelims : CharSet) : string;
    {-Given a set of word delimiters, return the N'th word in S}
  var
    I : Word;                 {!!.12}
    Count, Len : Byte;
    SLen : Byte absolute S;
  begin
    Count := 0;
    I := 1;
    Len := 0;
    ExtractWord[0] := #0;

    while (I <= SLen) and (Count <> N) do begin
      {skip over delimiters}
      while (I <= SLen) and (S[I] in WordDelims) do
        Inc(I);

      {if we're not beyond end of S, we're at the start of a word}
      if I <= SLen then
        Inc(Count);

      {find the end of the current word}
      while (I <= SLen) and not(S[I] in WordDelims) do begin
        {if this is the N'th word, add the I'th character to Tmp}
        if Count = N then begin
          Inc(Len);
          ExtractWord[0] := Char(Len);
          ExtractWord[Len] := S[I];
        end;

        Inc(I);
      end;
    end;
  end;

constructor TLinks.Create;
begin
  Demo := PrefixSeg <> 0;

  Active := True;
  with Screen do
    SavedRectangle := Rect(0, 0, Width, Height);

  Timer := TTimer.Create(Application);
  Timer.OnTimer := TimerElapsed;

  Update;
end;

destructor TLinks.Destroy;
begin
  StoreOptions;

  Active := False;
  Update{=Stop timer, Unhook};

  inherited Destroy;
end;


function TLinks.ComponentRect;
var
  L: Longint;
begin
  if Component is TControl
  then
    with TControl(Component), ClientToScreen(Point(0, 0)) do
      Result := Bounds(X, Y, Width, Height)
  else
  begin
    L := Component.DesignInfo;
    with TPoint(L) do
      Result := Bounds(X, Y, IconSize, IconSize);
    with HintForm, ClientToScreen(Point(0,0)) do
      OffsetRect(Result, X, Y);
  end;
end;


procedure TLinks.Update;
begin
  Cancel;

  if Active
  then
  begin
    Hook;
    Timer.Enabled := True;
  end
  else
  begin
    Timer.Enabled := False;
    UnHook;
  end;
end;

procedure TLinks.SetTemporaryDisabled;
begin
  FTemporaryDisabled := Value;

  if TemporaryDisabled then
    Cancel;
end;


function SomeKeyDown: Boolean;
var
  KeyState: TKeyboardState;
  I: Integer;
begin
  GetKeyboardState(KeyState);

  for I := Low(KeyState) to High(KeyState) do
    if KeyState[I] and $80 <> 0 then
    begin
      Result := True;
      Exit;
    end;

  Result := False;
end;

var
  Hooked: Boolean;

function MouseProc(Code: Integer; Msg: Word; lParam: Longint): Longint; export;
begin
  if Code >= 0 then
    with Links do
      case Msg of
        WM_LBUTTONDOWN,
        WM_RBUTTONDOWN:
          TemporaryDisabled := True;

        WM_LBUTTONUP,
        WM_RBUTTONUP:
          TemporaryDisabled := False;

        WM_LBUTTONDBLCLK:
          Cancel;
      end;

  Result := CallNextHookEx(Links.MouseHook, Code, Msg, lParam);
end;

function KeyboardProc(Code: Integer; Key: Word; KeyData: Longint): Longint; export;
begin
  if Code >= 0 then
    Links.TemporaryDisabled := SomeKeyDown;

  Result := CallNextHookEx(Links.KeyboardHook, Code, Key, KeyData)
end;

function TLinks.WindowHook;
begin
  Links.Cancel;

  Result := False;
end;

procedure TLinks.HookUnhookMainWindow;

  function DelphiApplication: HWnd;
  begin
    Result := FindWindow('TApplication', 'Delphi');
  end;

begin
  if not Demo{!} then
    SendMessage(DelphiApplication, CM_WINDOWHOOK, wParam, Longint(@@Hook));
end;

procedure TLinks.Hook;
begin
  if {}Hooked then
    Exit;
  Hooked := True;

  MouseHook := SetWindowsHookEx(WH_MOUSE, MouseProc,
    GetInstanceModule(HInstance), GetCurrentTask);
  KeyboardHook := SetWindowsHookEx(WH_KEYBOARD, KeyboardProc,
    GetInstanceModule(HInstance), GetCurrentTask);

  HookUnhookMainWindow(WindowHook, 0);
end;

procedure TLinks.UnHook;
begin
  if {}not Hooked then
    Exit;
  Hooked := False;

  HookUnhookMainWindow(WindowHook, 1);

  UnhookWindowsHookEx(KeyboardHook);
  UnhookWindowsHookEx(MouseHook);
end;

function TLinks.ComponentBelongsToForm;
var
  I: Integer;
begin
  with HintForm do
    for I := 0 to Pred(ComponentCount) do
      if Components[I] = Component then
      begin
        Result := True;
        Exit;
      end;

  Result := False;
end;

procedure TLinks.TimerElapsed;

  function FindForm(P: TPoint): TForm;
  var
    I: Integer;
    R: TRect;
  begin
    with Screen do
      for I := 0 to Pred(FormCount) do
      begin
        Result := Forms[I];
        R := Result.BoundsRect;
        if PtInRect(R, P) then
          Exit;
      end;

    Result := nil;
  end;

  function FindComponent(P: TPoint): TComponent;
  var
    I: Integer;
    Component: TComponent;
    R: TRect;
  begin
    Result := nil;
    with HintForm do
    begin
      for I := 0 to Pred(ComponentCount) do
      begin
        Component := Components[I];
        if
          (Component is TControl)  or
          (Component.DesignInfo = 0
            { ,    TxxxField}) then
          Continue;

        if PtInRect(ComponentRect(Component), P) then
        begin
          Result := Component;
          Exit;
        end;
      end;

      Result := FindDragTarget(P, True);
      if not Assigned(Result) then
        Exit;

      if Result = HintForm then
      begin
        Result := nil;
        Exit;
      end;

      if ComponentBelongsToForm(Result) then
        Exit{ };

      { TRadioGroup, }
      repeat
        if TControl(Result).Parent = HintForm then
          Exit;

        Result := TControl(Result).Parent;
      until False;
    end;
  end;

var
  MouseForm: TForm;
  MouseComponent: TComponent;
  P: TPoint;
begin
  if TemporaryDisabled then
    Exit;

  LoadOptions;

  GetCursorPos(P);
  MouseForm := FindForm(P);
  if Assigned(SavedScreen) then
    if MouseForm <> HintForm then
    begin
      Cancel;
      Exit;
    end;

  HintForm := MouseForm;
  if not Assigned(HintForm) then
    Exit;

  if not (Active and HintForm.Active) then
    Exit;

  if not Demo and not (csDesigning in HintForm.ComponentState){
    ,     ,     property editor'
     - .    -  - /  :(} then
    Exit;

  MouseComponent := FindComponent(P);
  if Assigned(SavedScreen)
  then
    if MouseComponent <> HintComponent
    then Cancel
    else {}
  else
    if Assigned(MouseComponent) then
      ActivateHint(MouseComponent);
end;


procedure TLinks.ActivateHint;
var
  DesktopCanvas: TCanvas;

  procedure SaveScreen;
  begin
    SavedScreen := TBitmap.Create;
    with SavedScreen do
    begin
      Width := SavedRectangle.Right - SavedRectangle.Left;
      Height := SavedRectangle.Bottom - SavedRectangle.Top;

      Canvas.CopyRect(Rect(0, 0, Width, Height), DesktopCanvas, SavedRectangle);
    end;
  end;

  procedure DrawInvisibleComponents;
  var
    I: Integer;
    Component: TComponent;
  begin
    with HintForm do
      for I := 0 to Pred(ComponentCount) do
      begin
        Component := Components[I];
        if
          not (Component is TControl){, ...} and
          (Component.DesignInfo <> 0{...    ,
             , ,   TxxxField}) then
        begin
          DesktopCanvas.FillRect(ComponentRect(Component));
        end;
      end;
  end;

  procedure DrawComponentRect(Component: TComponent; Color: TColor);
  begin
    with DesktopCanvas do
    begin
      Brush.Style := bsClear;
      Pen.Width := 3;
      Pen.Color := Color;
      with ComponentRect(Component) do
        Rectangle(Left, Top, Right, Bottom);

      Brush.Style := bsSolid;
      Pen.Width := 1;
    end;
  end;

  procedure DrawParentRect;
  begin
    with DesktopCanvas do
      if HintComponent is TControl then
        with TControl(HintComponent) do
          if (Parent <> nil) and (Parent <> HintForm) then
            DrawComponentRect(Parent, clTeal);
  end;

  procedure DrawSelfRect;
  begin
    DrawComponentRect(HintComponent, clBlack);
  end;

  procedure DrawHint;
  const
    MaxHintSize = $400;
  var
    HintWinRect: TRect;
    Hint: PChar;
  begin
    with DesktopCanvas do
    begin
      Brush.Color := HintColor;
      SetBkMode(Handle, TRANSPARENT);

      Hint := GetHint(HintComponent);
      Hint[Pred(Pred(StrLen(Hint)))] := #0;

      with ComponentRect(HintComponent) do
        HintWinRect := Bounds(
          Left, Bottom+2,
          Screen.Width-Left -3{     }, 0);

      DrawText(Handle, Hint, -1, HintWinRect,
        DT_CALCRECT or DT_LEFT or DT_WORDBREAK or DT_NOPREFIX);

      Inc(Longint(HintWinRect.BottomRight), $00020006);

      with HintWinRect do
        Rectangle(Left, Top, Right, Bottom);

      Inc(Longint(HintWinRect.TopLeft), $00010003);
      DrawText(Handle, Hint, -1, HintWinRect,
        DT_LEFT or DT_NOPREFIX or DT_WORDBREAK);

      Hint[StrLen(Hint)] := 'z';
      StrDispose(Hint);
    end;
  end;

  procedure DrawLinks;

    procedure DrawLink(Master, Slave: TComponent; Color: TColor);

      procedure ResetColor;
      begin
        with DesktopCanvas do
        begin
          Pen.Color := Color;
          Brush.Color := Color;
        end;
      end;

      procedure DrawLinkHandle(Component: TComponent; P: TPoint);
      begin
        with DesktopCanvas, P do
        begin
          if Component = HintComponent then
            Brush.Color := SelfColor;

          Ellipse(
            X-CircleSize, Y-CircleSize,
            Succ(X+CircleSize), Succ(Y+CircleSize));
        end;
      end;

    var
      MasterPoint, SlavePoint: TPoint;
    begin
      with DesktopCanvas do
      begin
{        with ComponentRect(Master) do
          MasterPoint := Point((Left+Right) div 2, (Top+Bottom) div 2);
        SlavePoint := ComponentRect(Slave).TopLeft;}

        MasterPoint := ComponentRect(Master).TopLeft;
        with ComponentRect(Slave) do
          SlavePoint := Point((Left+Right) div 2, (Top+Bottom) div 2);

(*        Dec(LongInt(MasterPoint), $00010001) {.BottomRight,   __  - };*)

        ResetColor;
        Pen.Width := LineWidth;
        PolyLine([MasterPoint, SlavePoint]);
        Pen.Width := 1;

        DrawLinkHandle(Master, MasterPoint);  ResetColor;
        DrawLinkHandle(Slave, SlavePoint);
      end;
    end;

  var
    Structure: TStrings;
    Unique: TList;

    procedure Recource(Component: TComponent; Up: Boolean);
    var
      I: Integer;
      Link: TComponent;
    begin
      with Structure do
        with TStrings(Objects[IndexOf(Component.Name)]){ Component'} do
          for I := 0 to Pred(Count) do
            if (Strings[I] = SSlave) xor Up then
            begin
              Link := TComponent(Objects[I]);

              if Up
              then DrawLink(Link, Component, MasterColor)
              else DrawLink(Component, Link, SlaveColor);

              if Unique.IndexOf(Link) >= 0 then
                Continue{ };
              Unique.Add(Link);

              Recource(Link, Up);
            end;
    end;

  var
    Links: TStrings;
    I: Integer;
    Component: TComponent;
  begin
    Unique := TList.Create;

    Structure := TStringList.Create;

    with HintForm do
    begin
      for I := 0 to Pred(ComponentCount) do
        Structure.AddObject(Components[I].Name, TStringList.Create);

      for I := 0 to Pred(ComponentCount) do
        CollectLinks(Structure, Components[I]);
    end;

    Recource(HintComponent, True);
    Unique.Clear;
    Recource(HintComponent, False);

    with Structure do
    begin
      for I := 0 to Pred(Count) do
        TObject(Objects[I]).Free;
      Free;
    end;

    Unique.Free;
  end;

begin
  HintComponent := AHintComponent;

  DesktopCanvas := TCanvas.Create;
  with DesktopCanvas do
  begin
    Handle := GetWindowDC(GetDesktopWindow);

    Font.Name := 'MS Sans Serif';
    Font.Size := 8;
  end;

  SaveScreen;

  if Demo then
    DrawInvisibleComponents;

  if ShowParentRect then DrawParentRect;
  if ShowSelfRect then DrawSelfRect;
  if ShowLinks then DrawLinks;
  if ShowHint then DrawHint;

  ReleaseDC(GetDesktopWindow, DesktopCanvas.Handle);
  DesktopCanvas.Free;
end;

procedure TLinks.Cancel;
var
  DesktopCanvas: TCanvas;

  procedure RestoreScreen;
  begin
    with SavedScreen do
      DesktopCanvas.CopyRect(SavedRectangle, Canvas, Rect(0, 0, Width, Height));
  end;

begin
  if not Assigned(SavedScreen) then
    Exit;

  DesktopCanvas := TCanvas.Create;
  with DesktopCanvas do
  begin
    Handle := GetWindowDC(GetDesktopWindow);

    RestoreScreen;

    ReleaseDC(GetDesktopWindow, Handle);
    Free;
  end;

  SavedScreen.Free;
  SavedScreen := nil;
end;

procedure TLinks.CollectLinks;

  function PropertyByName(const PropName: string): Pointer;
  var
    PropInfo: PPropInfo;
  begin
    Result := nil;

    PropInfo := GetPropInfo(Component.ClassInfo, PropName);
    if
      Assigned(PropInfo){  } and
      (PropInfo^.PropType^.Kind = tkClass{ }) then
      Result := Pointer(GetOrdProp(Component, PropInfo));
  end;

  procedure DoCollectLinks(const Props: string; IsMaster: Boolean);
  var
    N: Integer;
    Link: TComponent;
    Master, Slave: TComponent;
    S: string;
  begin
    for N := 1 to MaxInt do
    begin
      S := ExtractWord(N, Props, ListDelims);
      if S = '' then
        Break;

      Link := PropertyByName(S);
      if not Assigned(Link) or not ComponentBelongsToForm(Link) then
        Continue;

      if IsMaster
      then Master := Link
      else Master := Component;
      if IsMaster
      then Slave := Component
      else Slave := Link;

      with Structure do
      begin
        TStrings(Objects[IndexOf(Master.Name)]).AddObject(SSlave, Slave);
        TStrings(Objects[IndexOf(Slave.Name)]).AddObject(SMaster, Master);
      end;
    end;
  end;

begin
  DoCollectLinks(MasterProps, True);
  DoCollectLinks(SlaveProps, False);
end;

function TLinks.GetHint;

  function GetObjectText: TStrings;
  var
    Binary, Text: TMemoryStream;
    Writer: TWriter;
  begin
    Result := nil;

    try
      Binary := TMemoryStream.Create;
      Text := TMemoryStream.Create;
      try
        Writer := TWriter.Create(Binary, 4096);
        try
          Writer.Root := HintForm;
          Writer.WriteSignature;
          Writer.WriteComponent(Component);
        finally
          Writer.Free;
        end;
        Binary.Position := 0;

        ObjectBinaryToText(Binary, Text);
        Text.Position := 0;

        Result := TStringList.Create;
        Result.LoadFromStream(Text);
      finally
        Text.Free;
        Binary.Free;
      end;
    except
      { -  }
    end;
  end;

  procedure ProcessObjectText(Strings: TStrings);
  var
    InBinary: Boolean;

    function Process(var S: string): Boolean;

      function IsHidden: Boolean;
      var
        N: Integer;
        HiddenProp, HiddenPropStart: string;
      begin
        for N := 1 to MaxInt do
        begin
          HiddenProp := ExtractWord(N, HiddenProps, ListDelims);
          if HiddenProp = '' then
            Break;

          HiddenPropStart := HiddenProp+' = ';
          if Copy(S, 1, Length(HiddenPropStart))=HiddenPropStart then
          begin
            Result := True;
            Exit;
          end;
        end;

        Result := False;
      end;

      function IsBinary: Boolean;
      begin
        if Pos('{', S) <> 0 then
        begin
          InBinary := True;

          AppendStr(S, '...}');
          Result := False;
          Exit;
        end;

        Result := InBinary;

        if Pos('}', S) <> 0 then
          InBinary := False;
      end;

    begin
      Result := IsHidden or IsBinary;
    end;

  const
    ObjectStart = 'object ';
  var
    I: Integer;
    S: string;
    AfterLastNotEvent: Integer;
  begin
    with Strings do
    begin
      Strings[0] := Copy(Strings[0], Succ(Length(ObjectStart)), $FF){ 'object'};
      Delete(Pred(Count)){ 'end'};

      InBinary := False;
      I := 1;
      while I < Count do
      begin
        S := Copy(Strings[I], 3, $FF){   };

        if Copy(S, 1, Length(ObjectStart)) = ObjectStart{ -,
               ,  } then
        begin
          repeat
            Delete(I);
          until I = Count;
          Break;
        end;

        if Process(S)
        then
          Delete(I)
        else
        begin
          Strings[I] := S;  Inc(I);
        end;
      end;

      I := 1;
      AfterLastNotEvent := Count;
      while I < AfterLastNotEvent do
      begin
        if CompareText(Copy(Strings[I], 1, 2), 'ON') = 0 then
        begin
          Move(I, Pred{} (Count));  Dec(AfterLastNotEvent);
          Continue;
        end;

        Inc(I);
      end;
      if AfterLastNotEvent <> Count{-  } then
        Insert(AfterLastNotEvent, '');

      if Count <> 1{-  } then
        Insert(1, '');
    end;
  end;

var
  Strings: TStrings;
begin
  Strings := GetObjectText;

  if not Assigned(Strings){    } then
  begin
    Result := StrNew('Bug in .GetHint - report to author!'^M^J);
    Exit;
  end;

  ProcessObjectText(Strings);
  Result := Strings.GetText;

  Strings.Free;
end;


procedure TLinks.StoreOptions;
begin
  with TIniFile.Create(IniFileName) do
    try
      WriteInteger(SectionName, ICircleSize, CircleSize);
      WriteInteger(SectionName, ILineWidth, LineWidth);
      WriteString(SectionName, IMasterColor, ColorToString(MasterColor));
      WriteString(SectionName, ISelfColor, ColorToString(SelfColor));
      WriteString(SectionName, ISlaveColor, ColorToString(SlaveColor));
      WriteString(SectionName, IHintColor, ColorToString(HintColor));
      WriteInteger(SectionName, IHintPause, Timer.Interval);
      WriteString(SectionName, IMasterProps, MasterProps);
      WriteString(SectionName, ISlaveProps, SlaveProps);
      WriteString(SectionName, IHiddenProps, HiddenProps);
      WriteInteger(SectionName, IShowParentRect, Longint(ShowParentRect));
      WriteInteger(SectionName, IShowSelfRect, Longint(ShowSelfRect));
      WriteInteger(SectionName, IShowLinks, Longint(ShowLinks));
      WriteInteger(SectionName, IShowHint, Longint(ShowHint));
    finally
      Free;
    end;
end;

procedure TLinks.LoadOptions;

  function IniFileUnchanged: Boolean;
  var
    Buffer: array[0..$100] of Char;
    IniFileAge: Longint;
  begin
    GetWindowsDirectory(Buffer, SizeOf(Buffer));
    IniFileAge := FileAge(Format('%s\%s', [StrPas(Buffer), IniFileName]));

    Result := IniFileAge = PrevIniFileAge;

    PrevIniFileAge := IniFileAge;
  end;

begin
  if IniFileUnchanged then
    Exit;

  with TIniFile.Create(IniFileName) do
    try
      CircleSize := ReadInteger(SectionName, ICircleSize, DefaultCircleSize);
      LineWidth := ReadInteger(SectionName, ILineWidth, DefaultLineWidth);
      MasterColor := StringToColor(ReadString(SectionName, IMasterColor, ColorToString(DefaultMasterColor)));
      SelfColor := StringToColor(ReadString(SectionName, ISelfColor, ColorToString(DefaultSelfColor)));
      SlaveColor := StringToColor(ReadString(SectionName, ISlaveColor, ColorToString(DefaultSlaveColor)));
      HintColor := StringToColor(ReadString(SectionName, IHintColor, ColorToString(DefaultHintColor)));
      Timer.Interval := ReadInteger(SectionName, IHintPause, DefaultHintPause);
      MasterProps := ReadString(SectionName, IMasterProps, DefaultMasterProps);
      SlaveProps := ReadString(SectionName, ISlaveProps, DefaultSlaveProps);
      HiddenProps := ReadString(SectionName, IHiddenProps, DefaultHiddenProps);
      ShowParentRect := LongBool(ReadInteger(SectionName, IShowParentRect, Longint(True)));
      ShowSelfRect := LongBool(ReadInteger(SectionName, IShowSelfRect, Longint(True)));
      ShowLinks := LongBool(ReadInteger(SectionName, IShowLinks, Longint(True)));
      ShowHint := LongBool(ReadInteger(SectionName, IShowHint, Longint(True)));
    finally
      Free;
    end;
end;

procedure ExitProc; far;{,     complib   $F+,
      default}
begin
  Links.Free;
end;

initialization
  Links := TLinks.Create;
  AddExitProc(ExitProc);
end.

